/*
 * Grundkurs Informatik - Beispielprogramm zu direkten Sortierverfahren
 * J. Schmidt, H. Ernst 2016
 *
 * direktes Einfügen, direktes binäres Einfügen, direktes Auswählen, Bubblesort und Shakersort
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

// Definition von bool für C90
typedef enum {false=0, true=1} bool;

// sortiert werden hier double-Zahlen, die in einem Array gespeichert sind
// Arraygröße
#define ANZ 10000

// Datenarrays
// diese wurden global angelegt, da bei lokaler Deklaration in main()
// bei großen Arrays ein Stacküberlauf eintreten wird
double daten[ANZ];
double datenkopie[ANZ];

// Initalisierung des Datenarrays mit Zufallszahlen
// Parameter: Datenarray, Größe des Arrays, maximale Größe der double Zahle
void initArray(double daten[], size_t n, double maxBetrag) {
    size_t i;

    srand((unsigned) time(0)); // init. Zufallszahlengenerator

    for(i = 0; i < n; i++) {
        daten[i] = ((double) rand() / (double) RAND_MAX) - 0.5;  // doubles zwischen -0.5 und +0.5
        daten[i] = 2.0 * daten[i] * maxBetrag;  // strecken auf den gesamten Bereich
    }
}

// Kopie eines double-Arrays anfertigen ziel = quelle
void copyDoubleArray(double ziel[], double quelle[], size_t n) {
    memcpy(ziel, quelle, n * sizeof(double));
}

// erste n Werte des Datenarrays ausgeben
void printDaten(double daten[], size_t n) {
    size_t i;

    printf("Die ersten %ld Datenwerte:\n", n);
    for(i = 0; i < n; i++) {
        printf("%f ", daten[i]);
    }
    printf("\n");
}


// direktes Einfügen - Insertion Sort
void insertionSort(double daten[], size_t n) {
    int i, j;  // Zähler
    double x;     // aktuelles Element

    for(i = 1; i < n; i++) {
        // nächstes Element auswählen
        x = daten[i];
        j = i - 1;

        // nach rechts verschieben
        while(j >= 0 && x < daten[j]) {
            daten[j+1] = daten[j];
            j--;
        }

        daten[j+1] = x;
    }
}

// direktes binäres Einfügen - Insertion Sort (binary)
void insertionSortBin(double daten[], size_t n) {
    int i, j;  // Zähler
    int ug, og;   // Ober-/Untergrenze
    double x;     // aktuelles Element

    for(i = 1; i < n; i++) {
        x = daten[i];
        ug= 0;
        og= i - 1;

        while(ug <= og) {
            j = (ug + og) / 2;
            if(x < daten[j])
                og = j - 1;
            else
                ug = j + 1;
        }

        for(j = i; j > ug; j--)
            daten[j] = daten[j-1];

        daten[ug]=x;
    }
}

// direktes Auswählen - Selection Sort
void selectionSort(double daten[], size_t n) {
    int i, j, k;  // Zähler
    double x;     // aktuelles Element

    for(i = 0; i < n-1; i++) { // Array durchlaufen
        x = daten[i];
        k = i;
        // Minimum suchen
        for(j = i+1; j < n; j++)
            if(daten[j] < x) {
                k = j;
                x = daten[j];
            }
        // Austauschen
        daten[k] = daten[i];
        daten[i]=x;
    }
}

// Bubble Sort
void bubbleSort(double daten[], size_t n) {
    int i, j;  // Zähler
    int k, kstrich;     //  Index des letzten Austauschs
    double x;  // aktuelles Element
    bool flag;  // true, wenn keine Veränderung mehr auftritt

    k = kstrich = 1;
    for(i = 1; i < n; i++) {
        flag = true;

        for(j = n-1; j >= k; j--)
            if(daten[j-1] > daten[j]) { // Austauschen
                x = daten[j-1];
                daten[j-1] = daten[j];
                daten[j] = x;
                flag = false;
                kstrich = j;   // Index merken
            }

        if(flag == true)
            return; // Fertig

        k = kstrich;
    }
}

// Shakersort
void shakerSort(double daten[], size_t n) {
    int i, j;  // Zähler
    double x;  // aktuelles Element
    int ug, og;  // Ober-/Untergrenze
    bool flag;  // true, wenn keine Veränderung mehr auftritt

    ug = 1;
    og = n;

    do {
        flag = true;

        for(i = ug; i < og; i++)  // Nach oben arbeiten
            if(daten[i-1] > daten[i]) { // Austauschen
                x = daten[i-1];
                daten[i-1] = daten[i];
                daten[i] = x;
                flag = false;
            }
        if(flag == true) return; // Fertig

        flag = true;
        for(i = --og - 1; i >= ug; i--) // Nach unten arbeiten
            if(daten[i-1] > daten[i]) {   // Austauschen
                x = daten[i-1];
                daten[i-1] = daten[i];
                daten[i] = x;
                flag = false;
            }

        if(flag == true) return; // Fertig
        ug++;
    } while(flag == false);
}

int main() {
    time_t t; // Zeit

    initArray(daten, ANZ, 50.0);
    copyDoubleArray(datenkopie, daten, ANZ);  // Unsortiertes Array merken für spätere Verwendung
    printDaten(daten, 20);

    printf("\nInsertion Sort\n");
    t = clock();
    insertionSort(daten, ANZ);
    printf("beendet nach %5.3f sec\n", // benötigte Zeit
           (float) (difftime(clock(), t) / CLOCKS_PER_SEC));
    printDaten(daten, 20);

    copyDoubleArray(daten, datenkopie, ANZ);  // Ursprungszustand (unsortiert) wieder herstellen
    t = clock();
    printf("\nInsertion Sort (binär)\n");
    insertionSortBin(daten, ANZ);
    printf("beendet nach %5.3f sec\n", // benötigte Zeit
           (float) (difftime(clock(), t) / CLOCKS_PER_SEC));
    printDaten(daten, 20);

    copyDoubleArray(daten, datenkopie, ANZ);  // Ursprungszustand (unsortiert) wieder herstellen
    t = clock();
    printf("\nSelection Sort\n");
    selectionSort(daten, ANZ);
    printf("beendet nach %5.3f sec\n", // benötigte Zeit
           (float) (difftime(clock(), t) / CLOCKS_PER_SEC));
    printDaten(daten, 20);

    copyDoubleArray(daten, datenkopie, ANZ);  // Ursprungszustand (unsortiert) wieder herstellen
    t = clock();
    printf("\nBubble Sort\n");
    bubbleSort(daten, ANZ);
    printf("beendet nach %5.3f sec\n", // benötigte Zeit
           (float) (difftime(clock(), t) / CLOCKS_PER_SEC));
    printDaten(daten, 20);

    copyDoubleArray(daten, datenkopie, ANZ);  // Ursprungszustand (unsortiert) wieder herstellen
    t = clock();
    printf("\nShaker Sort\n");
    shakerSort(daten, ANZ);
    printf("beendet nach %5.3f sec\n", // benötigte Zeit
           (float) (difftime(clock(), t) / CLOCKS_PER_SEC));
    printDaten(daten, 20);

    return EXIT_SUCCESS;
}